import pytest


text_modules_data = [
    {
        "id": "test_without_localization",
        "header": "test header",
        "body": "test body",
    },
    {
        "id": "test_with_localization",
        "localizedHeader": {
            "defaultValue": {
                "language": "en",
                "value": "test header",
            },
            "translatedValues": [
                {
                    "language": "de",
                    "value": "Test Kopfzeile",
                }
            ],
        },
        "localizedBody": {
            "defaultValue": {
                "language": "en",
                "value": "test body",
            },
            "translatedValues": [
                {
                    "language": "de",
                    "value": "Test Textkörper",
                }
            ],
        },
    },
]

params_for_create = [
    (
        "Generic",
        {
            "textModulesData": text_modules_data,
        },
        {
            "state": "INACTIVE",
            "cardTitle": {
                "defaultValue": {
                    "language": "en",
                    "value": "test card title",
                },
            },
            "header": {
                "defaultValue": {
                    "language": "en",
                    "value": "test header",
                },
            },
        },
    ),
    (
        "GiftCard",
        {
            "issuerName": "test issuer",
            "reviewStatus": "DRAFT",
            "textModulesData": text_modules_data,
        },
        {
            "state": "INACTIVE",
            "cardNumber": "test card number 1234",
        },
    ),
    (
        "Loyalty",
        {
            "issuerName": "test issuer",
            "programName": "test program",
            "programLogo": {
                "sourceUri": {
                    "uri": "https://edutap.eu/static/3d642ea43b0bc1fb150510749f6c2333/084c2/eugloh-white.webp",
                },
                "contentDescription": {
                    "defaultValue": {
                        "language": "en",
                        "value": "test logo",
                    },
                },
            },
            "reviewStatus": "DRAFT",
            "textModulesData": text_modules_data,
        },
        {
            "state": "INACTIVE",
        },
    ),
    (
        "Offer",
        {
            "issuerName": "test issuer",
            "reviewStatus": "DRAFT",
            "textModulesData": text_modules_data,
            "title": "test title",
            "redemptionChannel": "ONLINE",
            "provider": "test provider",
        },
        {
            "state": "INACTIVE",
        },
    ),
]


@pytest.mark.integration
@pytest.mark.parametrize("type_base,class_data,object_data", params_for_create)
def test_class_object_cru(type_base, class_data, object_data, integration_test_id):
    from edutap.wallet_google.api import create
    from edutap.wallet_google.api import listing
    from edutap.wallet_google.api import message
    from edutap.wallet_google.api import new
    from edutap.wallet_google.api import read
    from edutap.wallet_google.api import session_manager
    from edutap.wallet_google.api import update
    from edutap.wallet_google.registry import lookup_metadata_by_name

    import time

    class_type = f"{type_base}Class"
    object_type = f"{type_base}Object"

    ############################
    # test class
    class_base = f"{integration_test_id}.{class_type}.test_CRU.wallet_google.edutap"
    class_data["id"] = f"{session_manager.settings.test_issuer_id}.{class_base}"

    data = new(class_type, class_data)

    # create
    result_create = create(data)
    assert result_create is not None
    assert result_create.textModulesData[0].id == class_data["textModulesData"][0]["id"]
    assert (
        result_create.textModulesData[0].header
        == class_data["textModulesData"][0]["header"]
    )

    # relax - not sure if this is necessary
    time.sleep(0.05)

    # read
    result_read = read(name=class_type, resource_id=class_data["id"])
    assert result_read is not None
    assert (
        result_read.textModulesData[0].body == class_data["textModulesData"][0]["body"]
    )
    if "reviewStatus" in class_data:
        assert result_read.reviewStatus == class_data["reviewStatus"]

    # update - full
    result_read.textModulesData[0].body = "updated body"
    if type_base != "Generic":
        # generic has no reviewState
        result_read.reviewStatus = "UNDER_REVIEW"
    result_updated = update(result_read)
    assert result_updated is not None
    assert result_updated.textModulesData[0].body == "updated body"

    # read after update
    result_read = read(name=class_type, resource_id=class_data["id"])
    assert result_read is not None
    assert result_read.textModulesData[0].body == "updated body"

    model_metadata = lookup_metadata_by_name(class_type)
    if model_metadata["can_message"]:
        # send message to all
        result_message = message(
            name=class_type,
            resource_id=class_data["id"],
            message={
                "messageType": "TEXT",
                "id": "test-message",
                "header": "test header",
                "body": "test body",
            },
        )
        assert result_message is not None
        assert result_message.id == class_data["id"]

    # list all
    result_list = [
        x
        for x in listing(
            name=class_type, issuer_id=session_manager.settings.test_issuer_id
        )
    ]
    assert len(result_list) > 0

    ############################
    # test object
    object_data["classId"] = class_data["id"]
    object_base = f"{integration_test_id}.{object_type}.test_CRU.wallet_google.edutap"
    object_data["id"] = f"{session_manager.settings.test_issuer_id}.{object_base}"
    odata = new(object_type, object_data)

    # create
    oresult_create = create(odata)
    assert oresult_create is not None
    assert oresult_create.state == "INACTIVE"

    # relax - not sure if this is necessary
    time.sleep(0.05)

    # read
    oresult_read = read(name=object_type, resource_id=object_data["id"])
    assert oresult_read is not None
    assert oresult_read.state == "INACTIVE"

    # update
    oresult_read.state = "ACTIVE"
    oresult_updated = update(oresult_read)
    assert oresult_updated is not None
    assert oresult_updated.state == "ACTIVE"

    # read after update
    oresult_read = read(name=object_type, resource_id=object_data["id"])
    assert oresult_read is not None
    assert oresult_read.state == "ACTIVE"

    # list all
    result_list = [x for x in listing(name=object_type, resource_id=class_data["id"])]
    assert len(result_list) == 1

    model_metadata = lookup_metadata_by_name(class_type)
    if model_metadata["can_message"]:
        # send message
        result_message = message(
            name=object_type,
            resource_id=object_data["id"],
            message={
                "messageType": "TEXT",
                "id": "test-message",
                "header": "test header",
                "body": "test body",
            },
        )
        assert result_message is not None
        assert result_message.id == object_data["id"]



from .registry import lookup_metadata_by_name
from .settings import Settings
from google.auth.transport.requests import AuthorizedSession
from google.oauth2.service_account import Credentials
from requests.adapters import HTTPAdapter

import json
import threading


_THREADLOCAL = threading.local()


class HTTPRecorder(HTTPAdapter):
    """Record the HTTP requests and responses to a file."""

    @property
    def settings(self) -> Settings:
        settings = getattr(self, "_settings", None)
        if settings is None:
            self._settings = Settings()
        return self._settings

    def send(self, request, *args, **kwargs):
        req_record = {
            "method": request.method,
            "url": request.url,
            "headers": dict(request.headers),
            "body": json.loads(request.body.decode("utf-8")),
        }
        target_directory = self.settings.record_api_calls_dir
        if target_directory.is_dir() and not target_directory.exists():
            target_directory.mkdir(parents=True)
        filename = f"{target_directory}/{request.method}-{request.url.replace('/', '_')}.REQUEST.json"
        with open(filename, "w") as fp:
            json.dump(req_record, fp, indent=4)
        response = super().send(request, *args, **kwargs)
        resp_record = {
            "status_code": response.status_code,
            "headers": dict(response.headers),
            "body": response.json(),
        }
        with open(filename.replace("REQUEST", "RESPONSE"), "w") as fp:
            json.dump(resp_record, fp, indent=4)
        return response


class SessionManager:
    """Manages the session to the Google Wallet API and provides helper methods.

    Sessions here are thread safe.
    """

    @property
    def settings(self) -> Settings:
        settings = getattr(self, "_settings", None)
        if settings is None:
            self._settings = Settings()
        return self._settings

    def _make_session(self) -> AuthorizedSession:
        if not self.settings.credentials_file.is_file():
            raise ValueError(
                f"EDUTAP_WALLET_GOOGLE_CREDENTIALS_FILE={self.settings.credentials_file} does not exist."
            )
        credentials = Credentials.from_service_account_file(
            str(self.settings.credentials_file),
            scopes=self.settings.credentials_scopes,
        )
        session = AuthorizedSession(credentials)
        if self.settings.record_api_calls_dir is not None:
            session.mount("https://", HTTPRecorder())
        return session

    @property
    def session(self) -> AuthorizedSession:
        if getattr(_THREADLOCAL, "session", None) is None:
            _THREADLOCAL.session = self._make_session()
        return _THREADLOCAL.session  # type: ignore

    def url(self, name: str, additional_path: str = "") -> str:
        """
        Create the URL for the CRUD operations.

        :param name:            Registered name of the model.
        :param additional_path: Append this to the path.
                                Must start with a forward slash.

        :return: the url of the google RESTful API endpoint to handle this model
        """
        model_metadata = lookup_metadata_by_name(name)
        return f"{self.settings.api_url}/{model_metadata['url_part']}{additional_path}"


session_manager = SessionManager()



from ..bases import Model
from ..deprecated import DeprecatedKindFieldMixin
from .enums import BoardingPolicy
from .enums import SeatClassPolicy
from .general import Image
from .general import LocalizedString
from pydantic import Field


# Attribute order as in Google's documentation to make future updates easier!
# last check: 2025-01-22


class FlightCarrier(DeprecatedKindFieldMixin, Model):
    """
    see: https://developers.google.com/wallet/reference/rest/v1/flightclass#flightcarrier
    """

    # inherits kind (deprecated)
    carrierIataCode: str | None = Field(
        max_length=2,
        default=None,
    )
    carrierIcaoCode: str | None = Field(
        max_length=3,
        default=None,
    )
    airlineName: LocalizedString | None = None
    airlineLogo: Image | None = None
    airlineAllianceLogo: Image | None = None
    wideAirlineLogo: Image | None = None


class FlightHeader(DeprecatedKindFieldMixin, Model):
    """
    see: https://developers.google.com/wallet/reference/rest/v1/flightclass#flightheader
    """

    # inherits kind (deprecated)
    carrier: FlightCarrier | None = None
    flightNumber: str | None = None
    operatingCarrier: FlightCarrier | None = None
    operatingFlightNumber: str | None = None
    flightNumberDisplayOverride: str | None = None


class AirportInfo(DeprecatedKindFieldMixin, Model):
    """
    see: https://developers.google.com/wallet/reference/rest/v1/flightclass#airportinfo
    """

    # inherits kind (deprecated)
    airportIataCode: str | None = Field(max_length=3, default=None)
    terminal: str | None = None
    gate: str | None = None
    airportNameOverride: LocalizedString | None = None


class BoardingAndSeatingPolicy(DeprecatedKindFieldMixin, Model):
    """
    see: https://developers.google.com/wallet/reference/rest/v1/flightclass#boardingandseatingpolicy
    """

    # inherits kind (deprecated)
    boardingPolicy: BoardingPolicy = BoardingPolicy.BOARDING_POLICY_UNSPECIFIED
    seatClassPolicy: SeatClassPolicy = SeatClassPolicy.SEAT_CLASS_POLICY_UNSPECIFIED



from edutap.wallet_google.models.datatypes import enums

import pytest


testdata = [
    (
        "Create",
        "GenericClass",
        "POST",
        {
            "id": "3388000000022141777.test54.ycc.edutap",
            "enableSmartTap": False,
            "multipleDevicesAndHoldersAllowedStatus": enums.MultipleDevicesAndHoldersAllowedStatus.STATUS_UNSPECIFIED,
            "viewUnlockRequirement": enums.ViewUnlockRequirement.VIEW_UNLOCK_REQUIREMENT_UNSPECIFIED,
        },
    ),
    (
        "Create",
        "GenericObject",
        "POST",
        {
            "id": "3388000000022141777.obj53.test.ycc.edutap",
            "classId": "3388000000022141777.test.ycc.edutap",
            "state": enums.State.ACTIVE,
        },
    ),
]


@pytest.mark.parametrize("prefix,name,method,checkdata", testdata)
def test_api_create(mock_request_response, prefix, name, method, checkdata):
    from edutap.wallet_google.api import create
    from edutap.wallet_google.api import new
    from edutap.wallet_google.registry import lookup_model_by_name
    from edutap.wallet_google.session import session_manager

    request_data = mock_request_response(
        f"{prefix}{name}", session_manager.url(name), method
    )
    data = new(name, request_data["request"]["body"])
    result = create(data)

    model = lookup_model_by_name(name)
    assert isinstance(result, model)
    for key, value in checkdata.items():
        assert getattr(result, key) == value



from ...registry import register_model
from ..datatypes.enums import ConcessionCategory
from ..datatypes.enums import ConfirmationCodeLabel
from ..datatypes.enums import FlightStatus
from ..datatypes.enums import GateLabel
from ..datatypes.enums import NotificationSettingsForUpdates
from ..datatypes.enums import PassengerType
from ..datatypes.enums import ReviewStatus
from ..datatypes.enums import RowLabel
from ..datatypes.enums import SeatLabel
from ..datatypes.enums import SectionLabel
from ..datatypes.enums import TicketStatus
from ..datatypes.enums import TransitType
from ..datatypes.enums import TripType
from ..datatypes.event import EventDateTime
from ..datatypes.event import EventReservationInfo
from ..datatypes.event import EventSeat
from ..datatypes.event import EventVenue
from ..datatypes.flight import AirportInfo
from ..datatypes.flight import BoardingAndSeatingPolicy
from ..datatypes.flight import FlightHeader
from ..datatypes.general import Image
from ..datatypes.general import Uri
from ..datatypes.localized_string import LocalizedString
from ..datatypes.money import Money
from ..datatypes.review import Review
from ..datatypes.transit import ActivationOptions
from ..datatypes.transit import ActivationStatus
from ..datatypes.transit import DeviceContext
from ..datatypes.transit import PurchaseDetails
from ..datatypes.transit import TicketLeg
from ..datatypes.transit import TicketRestrictions
from ..deprecated import DeprecatedAllowMultipleUsersPerObjectMixin
from ..deprecated import DeprecatedInfoModuleDataFieldMixin
from ..deprecated import DeprecatedKindFieldMixin
from ..deprecated import DeprecatedLocationsFieldMixin
from ..deprecated import DeprecatedVersionFieldMixin
from ..deprecated import DeprecatedWordMarkFieldMixin
from .bases import ClassModel
from .bases import CommonLogosMixin
from .bases import ObjectModel
from .bases import StyleableMixin
from pydantic import Field


# Attribute order as in Google's documentation to make future updates easier!
# last check: 2025-01-22


@register_model(
    "EventTicketClass",
    url_part="eventTicketClass",
    plural="eventTicketClasses",
)
class EventTicketClass(
    DeprecatedKindFieldMixin,
    DeprecatedVersionFieldMixin,
    DeprecatedAllowMultipleUsersPerObjectMixin,
    DeprecatedLocationsFieldMixin,
    DeprecatedInfoModuleDataFieldMixin,
    DeprecatedWordMarkFieldMixin,
    StyleableMixin,
    CommonLogosMixin,
    ClassModel,
):
    """
    see: https://developers.google.com/wallet/tickets/events/rest/v1/eventticketclass
    """

    # inherits kind (deprecated)
    eventName: LocalizedString | None = None
    eventId: str | None = Field(default=None, max_length=64)
    # inherits logo
    venue: EventVenue | None = None
    dateTime: EventDateTime | None = None
    confirmationCodeLabel: ConfirmationCodeLabel | None = None
    customConfirmationCodeLabel: LocalizedString | None = None
    seatLabel: SeatLabel | None = None
    customSeatLabel: LocalizedString | None = None
    rowLabel: RowLabel | None = None
    customRowLabel: LocalizedString | None = None
    sectionLabel: SectionLabel | None = None
    customSectionLabel: LocalizedString | None = None
    gateLabel: GateLabel | None = None
    customGateLabel: LocalizedString | None = None
    finePrint: LocalizedString | None = None
    # inherits classTemplateInfo
    # inherits id
    # inherits version (deprecated)
    issuerName: str | None = None
    # inherits messages
    # inherits allowMultipleUsersPerObject (deprecated)
    homepageUri: Uri | None = None
    # inherits locations (deprecated)
    reviewStatus: ReviewStatus | None = None
    review: Review | None = None
    # inherits infoModuleData (deprecated)
    # inherits imageModulesData
    # inherits textModulesData
    # inherits linksModuleData
    # inherits redemptionIssuers
    countryCode: str | None = None
    # inherits heroImage
    # inherits wordMark (deprecated)
    # inherits enableSmartTap
    # inherits hexBackgroundColor
    localizedIssuerName: LocalizedString | None = None
    # inherits multipleDevicesAndHoldersAllowedStatus
    # inherits callbackOptions
    # inherits securityAnimation
    # inherits viewUnlockRequirement
    # inherits wideLogo
    notifyPreference: NotificationSettingsForUpdates = (
        NotificationSettingsForUpdates.NOTIFICATION_SETTINGS_FOR_UPDATES_UNSPECIFIED
    )
    # inherits appLinkData
    # inherits valueAddedModuleData


@register_model("EventTicketObject", url_part="eventTicketObject")
class EventTicketObject(
    DeprecatedKindFieldMixin,
    DeprecatedVersionFieldMixin,
    DeprecatedLocationsFieldMixin,
    DeprecatedInfoModuleDataFieldMixin,
    StyleableMixin,
    CommonLogosMixin,
    ObjectModel,
):
    """
    see: https://developers.google.com/wallet/tickets/events/rest/v1/eventticketobject
    """

    # inherits kind (deprecated)
    classReference: EventTicketClass | None = None
    seatInfo: EventSeat | None = None
    reservationInfo: EventReservationInfo | None = None
    ticketHolderName: str | None = None
    ticketNumber: str | None = None
    ticketType: LocalizedString | None = None
    faceValue: Money | None = None
    # inherits groupingInfo
    linkedOfferIds: list[str] | None = None
    # inherits hexBackgroundColor
    # inherits id
    # inherits classId
    # inherits version (deprecated)
    # inherits state
    # inherits barcode
    # inherits messages
    # inherits validTimeInterval
    # inherits locations (deprecated)
    # inherits hasUsers
    # inherits smartTapRedemptionValue
    hasLinkedDevice: bool | None = False
    disableExpirationNotification: bool | None = False
    # inherits infoModuleData (deprecated)
    # inherits imageModulesData
    # inherits textModulesData
    # inherits linksModuleData
    # inherits appLinkData
    # inherits rotatingBarcode
    # inherits heroImage
    # inherits passConstraints
    # inherits saveRestrictions
    # inherits linkedObjectIds
    notifyPreference: NotificationSettingsForUpdates = (
        NotificationSettingsForUpdates.NOTIFICATION_SETTINGS_FOR_UPDATES_UNSPECIFIED
    )
    # inherits valueAddedModuleData


@register_model("TransitClass", url_part="transitClass", plural="transitClasses")
class TransitClass(
    DeprecatedVersionFieldMixin,
    DeprecatedAllowMultipleUsersPerObjectMixin,
    DeprecatedLocationsFieldMixin,
    DeprecatedInfoModuleDataFieldMixin,
    DeprecatedWordMarkFieldMixin,
    StyleableMixin,
    CommonLogosMixin,
    ClassModel,
):
    """
    see: https://developers.google.com/wallet/reference/rest/v1/transitclass
    """

    transitOperatorName: LocalizedString | None = None
    # inherits logo
    transitType: TransitType = TransitType.TRANSIT_TYPE_UNSPECIFIED
    watermark: Image | None = None
    languageOverride: str | None = None
    customTransitTerminusNameLabel: LocalizedString | None = None
    customTicketNumberLabel: LocalizedString | None = None
    customRouteRestrictionsLabel: LocalizedString | None = None
    customRouteRestrictionsDetailsLabel: LocalizedString | None = None
    customTimeRestrictionsLabel: LocalizedString | None = None
    customOtherRestrictionsLabel: LocalizedString | None = None
    customPurchaseReceiptNumberLabel: LocalizedString | None = None
    customConfirmationCodeLabel: LocalizedString | None = None
    customPurchaseFaceValueLabel: LocalizedString | None = None
    customPurchasePriceLabel: LocalizedString | None = None
    customDiscountMessageLabel: LocalizedString | None = None
    customCarriageLabel: LocalizedString | None = None
    customSeatLabel: LocalizedString | None = None
    customCoachLabel: LocalizedString | None = None
    customPlatformLabel: LocalizedString | None = None
    customZoneLabel: LocalizedString | None = None
    customFareClassLabel: LocalizedString | None = None
    customConcessionCategoryLabel: LocalizedString | None = None
    customFareNameLabel: LocalizedString | None = None
    # inherits classTemplateInfo
    enableSingleLegItinerary: bool = False
    # inherits id
    # inherits version (deprecated)
    issuerName: str | None = None
    # inherits messages
    # inherits allowMultipleUsersPerObject (deprecated)
    homepageUri: Uri | None = None
    # inherits locations (deprecated)
    reviewStatus: ReviewStatus | None = None
    review: Review | None = None
    # inherits infoModuleData (deprecated)
    # inherits imageModulesData
    # inherits textModulesData
    # inherits linksModuleData
    # inherits redemptionIssuers
    countryCode: str | None = None
    # inherits heroImage
    # inherits wordMark (deprecated)
    # inherits enableSmartTap
    # inherits hexBackgroundColor
    localizedIssuerName: LocalizedString | None = None
    # inherits multipleDevicesAndHoldersAllowedStatus
    # inherits callbackOptions
    # inherits securityAnimation
    activationOptions: ActivationOptions | None = None
    # inherits viewUnlockRequirement
    # inherits wideLogo
    notifyPreference: NotificationSettingsForUpdates = (
        NotificationSettingsForUpdates.NOTIFICATION_SETTINGS_FOR_UPDATES_UNSPECIFIED
    )
    # inherits appLinkData
    # inherits valueAddedModuleData


@register_model("TransitObject", url_part="transitObject")
class TransitObject(
    DeprecatedVersionFieldMixin,
    DeprecatedLocationsFieldMixin,
    DeprecatedInfoModuleDataFieldMixin,
    StyleableMixin,
    CommonLogosMixin,
    ObjectModel,
):
    """
    see: https://developers.google.com/wallet/reference/rest/v1/transitobject
    """

    classReference: TransitClass | None = None
    ticketNumber: str | None = None
    passengerType: PassengerType = PassengerType.PASSENGER_TYPE_UNSPECIFIED
    passengerNames: str | None = None
    tripId: str | None = None
    ticketStatus: TicketStatus = TicketStatus.TICKET_STATUS_UNSPECIFIED
    customTicketStatus: LocalizedString | None = None
    concessionCategory: ConcessionCategory = (
        ConcessionCategory.CONCESSION_CATEGORY_UNSPECIFIED
    )
    customConcessionCategory: LocalizedString | None = None
    # TODO: validator: Only one of the two fields should be set, i.e. either customConcessionCategory or concessionCategory
    ticketRestrictions: TicketRestrictions | None = None
    purchaseDetails: PurchaseDetails | None = None
    ticketLeg: TicketLeg | None = None
    ticketLegs: list[TicketLeg] | None = None
    # TODO: validator: If more than one leg is to be specified then use the ticketLegs field instead.
    #                  Both ticketLeg and ticketLegs may not be set.
    # inherits hexBackgroundColor
    tripType: TripType = TripType.TRIP_TYPE_UNSPECIFIED
    # inherits id
    # inherits classId
    # inherits version (deprecated)
    # inherits state
    # inherits barcode
    # inherits messages
    # inherits validTimeInterval
    # inherits locations (deprecated)
    # inherits hasUsers
    # inherits smartTapRedemptionValue
    hasLinkedDevice: bool | None = False
    disableExpirationNotification: bool | None = False
    # inherits infoModuleData (deprecated)
    # inherits imageModulesData
    # inherits textModulesData
    # inherits linksModuleData
    # inherits appLinkData
    activationStatus: ActivationStatus | None = None
    # inherits rotatingBarcode
    deviceContext: DeviceContext | None = None
    # inherits heroImage
    # inherits groupingInfo
    # inherits passConstraints
    # inherits saveRestrictions
    # inherits linkedObjectIds
    notifyPreference: NotificationSettingsForUpdates = (
        NotificationSettingsForUpdates.NOTIFICATION_SETTINGS_FOR_UPDATES_UNSPECIFIED
    )
    # inherits valueAddedModuleData


@register_model("FlightClass", url_part="flightClass", plural="flightClasses")
class FlightClass(
    DeprecatedKindFieldMixin,
    DeprecatedVersionFieldMixin,
    DeprecatedAllowMultipleUsersPerObjectMixin,
    DeprecatedLocationsFieldMixin,
    DeprecatedInfoModuleDataFieldMixin,
    DeprecatedWordMarkFieldMixin,
    StyleableMixin,
    CommonLogosMixin,
    ClassModel,
):
    """
    see: https://developers.google.com/wallet/reference/rest/v1/flightclass
    """

    # inherits kind (deprecated)
    localScheduledDepartureDateTime: str | None = Field(default=None)
    localEstimatedOrActualDepartureDateTime: str | None = Field(default=None)
    localBoardingDateTime: str | None = Field(default=None)
    localScheduledArrivalDateTime: str | None = Field(default=None)
    localEstimatedOrActualArrivalDateTime: str | None = Field(default=None)
    flightHeader: FlightHeader
    origin: AirportInfo
    destination: AirportInfo
    flightStatus: FlightStatus = FlightStatus.FLIGHT_STATUS_UNSPECIFIED
    boardingAndSeatingPolicy: BoardingAndSeatingPolicy | None = None
    localGateClosingDateTime: str | None = Field(default=None)
    # inherits classTemplateInfo
    languageOverride: str | None = None
    # inherits id
    # inherits version (deprecated)
    issuerName: str | None = None
    # inherits messages
    # inherits allowMultipleUsersPerObject (deprecated)
    # inherits homepageUri
    # inherits locations (deprecated)
    reviewStatus: ReviewStatus | None = None
    review: Review | None = None
    # inherits infoModuleData (deprecated)
    # inherits imageModulesData
    # inherits textModulesData
    # inherits linksModuleData
    # inherits redemptionIssuers
    countryCode: str | None = None
    # inherits heroImage
    # inherits wordMark (deprecated)
    # inherits enableSmartTap
    # inherits hexBackgroundColor
    localizedIssuerName: LocalizedString | None = None
    # inherits multipleDevicesAndHoldersAllowedStatus
    # inherits callbackOptions
    # inherits securityAnimation
    # inherits viewUnlockRequirement
    notifyPreference: NotificationSettingsForUpdates = (
        NotificationSettingsForUpdates.NOTIFICATION_SETTINGS_FOR_UPDATES_UNSPECIFIED
    )
    # inherits appLinkData
    # inherits valueAddedModuleData


@register_model("FlightObject", url_part="flightObject")
class FlightObject(
    DeprecatedKindFieldMixin,
    DeprecatedVersionFieldMixin,
    DeprecatedLocationsFieldMixin,
    StyleableMixin,
    ObjectModel,
):
    """
    see: https://developers.google.com/wallet/reference/rest/v1/flightobject
    """

    # inherits kind (deprecated)
    classReference: FlightClass | None = None
    passengerName: str | None = None
    # TODO: boardingAndSeatingInfo
    # TODO: reservationInfo
    securityProgramLogo: Image | None = None
    # inherits hexBackgroundColor
    # inherits id
    # inherits classId
    # inherits version (deprecated)
    # inherits state
    # inherits barcode
    # inherits messages
    # inherits validTimeInterval
    # inherits locations (deprecated)
    # inherits hasUsers
    # inherits smartTapRedemptionValue
    hasLinkedDevice: bool | None = False
    disableExpirationNotification: bool | None = False
    # inherits infoModuleData
    # inherits imageModulesData
    # inherits textModulesData
    # inherits linksModuleData
    # inherits appLinkData
    # inherits rotatingBarcode
    heroImage: Image | None = None
    # inherits groupingInfo
    # inherits passConstraints
    # inherits saveRestrictions
    # inherits linkedObjectIds
    notifyPreference: NotificationSettingsForUpdates = (
        NotificationSettingsForUpdates.NOTIFICATION_SETTINGS_FOR_UPDATES_UNSPECIFIED
    )
    # inherits valueAddedModuleData



import pytest


def test_camel_case_alias_enum():
    from edutap.wallet_google.models.bases import CamelCaseAliasEnum

    class TestFoo(CamelCaseAliasEnum):
        FOO_BAR_BAZ = "FOO_BAR_BAZ"

    assert TestFoo("FOO_BAR_BAZ") == TestFoo.FOO_BAR_BAZ
    assert TestFoo("fooBarBaz") == TestFoo.FOO_BAR_BAZ
    assert TestFoo("fooBarBaz") == "FOO_BAR_BAZ"
    with pytest.raises(ValueError):
        assert TestFoo("fooBarBaz") == "wrong"

    # test for https://github.com/edutap-eu/edutap.wallet_google/issues/12
    assert (
        repr(TestFoo("fooBarBaz")) == "<TestFoo.TestFoo camel case alias: 'fooBarBaz'>"
    )
    assert repr(TestFoo("FOO_BAR_BAZ")) == "<TestFoo.FOO_BAR_BAZ: 'FOO_BAR_BAZ'>"


def test_action_enum():
    from edutap.wallet_google.models.datatypes.enums import Action

    assert Action("ACTION_UNSPECIFIED") == Action.ACTION_UNSPECIFIED
    assert Action("actionUnspecified") == Action.ACTION_UNSPECIFIED

    assert Action("S2AP") != Action.ACTION_UNSPECIFIED

    assert Action("S2AP") == Action.S2AP
    assert Action("s2ap") == Action.S2AP

    assert Action("SIGN_UP") == Action.SIGN_UP
    assert Action("signUp") == Action.SIGN_UP


def test_pydantic_constraints():
    from edutap.wallet_google.models.bases import CamelCaseAliasEnum

    import pydantic

    class TestFoo(CamelCaseAliasEnum):
        FOO_BAR_BAZ = "FOO_BAR_BAZ"

    class TestModel(pydantic.BaseModel):
        foo: TestFoo = TestFoo.FOO_BAR_BAZ

    # accept snake case enum
    TestModel(foo=TestFoo("FOO_BAR_BAZ"))

    # accept snake case string
    TestModel(foo="FOO_BAR_BAZ")

    # accept camel case enum
    TestModel(foo=TestFoo("fooBarBaz"))

    # accept camel case string
    TestModel(foo="fooBarBaz")

    with pytest.raises(pydantic.ValidationError):
        TestModel(foo="wrong")



def test_encrypt_decrypt(mock_fernet_encryption_key):
    from edutap.wallet_google.utils import decrypt_data
    from edutap.wallet_google.utils import encrypt_data

    data = "hypoknapsenverdraller"
    encrypted_data = encrypt_data(data)
    assert encrypted_data != data
    assert data == decrypt_data(encrypted_data)


def test_generate_fernet_key(capsys):
    from edutap.wallet_google.utils import generate_fernet_key

    generate_fernet_key()
    captured = capsys.readouterr()
    assert captured.out.endswith("=\n")



from ..bases import Model
from .datetime import TimeInterval
from .general import Image
from .localized_string import LocalizedString


# Attribute order as in Google's documentation to make future updates easier!
# last check: 2025-01-22


class ModuleViewConstraints(Model):
    """
    see: https://developers.google.com/wallet/reference/rest/v1/ValueAddedModuleData
    """

    displayInterval: TimeInterval | None = None


class ValueAddedModuleData(Model):
    """
    see: https://developers.google.com/wallet/reference/rest/v1/ValueAddedModuleData
    """

    header: LocalizedString | None = None
    body: LocalizedString | None = None
    image: Image | None = None
    uri: str | None = None
    viewConstraints: ModuleViewConstraints | None = None
    sortIndex: int | None = None



# import to enable direkt usage import from parent namespace
# models also registers the models in the registry
from . import api  # noqa: F401
from . import models  # noqa: F401
from . import utils  # noqa: F401



from .models.bases import Model
from typing import TypedDict


class RegistryMetadataDict(TypedDict, total=False):
    """TypedDict for the metadata of a registered model."""

    model: type[Model]
    name: str
    url_part: str
    plural: str
    resource_id: str
    can_create: bool
    can_read: bool
    can_update: bool
    can_list: bool
    can_message: bool


_MODEL_REGISTRY_BY_NAME: dict[str, RegistryMetadataDict] = {}
_MODEL_REGISTRY_BY_MODEL: dict[type[Model], RegistryMetadataDict] = {}


class register_model:
    """
    Registers a Pydantic model based on Model in a registry.
    To be used as a decorator.
    """

    def __init__(
        self,
        name: str,
        url_part: str,
        *,
        plural: str | None = None,
        resource_id: str = "id",
        can_create: bool = True,
        can_read: bool = True,
        can_update: bool = True,
        can_list: bool = True,
        can_message: bool = True,
    ):
        """
        Prepares the registration of a Model or its subclasses.

        :param name:        Name of the model. Usually the same as the class name.
        :param url_part:    Part of the URL to be used for the RESTful API endpoint.
        :param plural:      Lowercase plural of the name. Defaults to url_part + 's'.
        :param resource_id: key name to fetch resource_id from. Defaults to 'id'.
        :param can_create:  Whether it is possible to use in the 'create' API function.
                            Defaults to True.
        :param can_read:    Whether it is possible to use in the 'read' API function.
                            Defaults to True.
        :param can_update:  Whether it is possible to use in the 'update' API function.
                            Defaults to True.
        :param can_list:    Whether it is possible to use in the 'list' API function.
                            Defaults to True.
        :param can_message: Whether it is possible to use in the 'message' API function.
                            Defaults to True.
        """
        self.metadata: RegistryMetadataDict = {
            "name": name,
            "url_part": url_part,
            "plural": plural or f"{url_part}s",
            "resource_id": resource_id,
            "can_create": can_create,
            "can_read": can_read,
            "can_update": can_update,
            "can_list": can_list,
            "can_message": can_message,
        }

    def __call__(
        self,
        cls: type[Model],
    ) -> type[Model]:
        """
        Registers the given class in the registry.
        """
        name = self.metadata["name"]
        if name in _MODEL_REGISTRY_BY_NAME:
            raise ValueError(f"Duplicate registration of '{name}'")
        self.metadata["model"] = cls
        _MODEL_REGISTRY_BY_NAME[name] = self.metadata
        _MODEL_REGISTRY_BY_MODEL[cls] = self.metadata
        return cls


def lookup_model_by_name(name: str) -> type[Model]:
    """
    Returns the model with the given name.
    """
    return _MODEL_REGISTRY_BY_NAME[name]["model"]


def lookup_model_by_plural_name(plural_name: str) -> type[Model]:
    """
    Returns the model with the given plural name.
    """
    for model in _MODEL_REGISTRY_BY_NAME.values():
        if model["plural"] == plural_name:
            return model["model"]
    raise LookupError(f"Model with plural name '{plural_name}' not found")


def lookup_metadata_by_name(name: str) -> RegistryMetadataDict:
    """
    Returns the metadata of the model with the given name.
    """
    return _MODEL_REGISTRY_BY_NAME[name]


def lookup_metadata_by_model_instance(model: Model) -> RegistryMetadataDict:
    """
    Returns the registry metadata by a given instance of a model
    """
    return _MODEL_REGISTRY_BY_MODEL[type(model)]


def lookup_metadata_by_model_type(model_type: type[Model]) -> RegistryMetadataDict:
    """
    Returns the registry metadata by a given model type
    """
    return _MODEL_REGISTRY_BY_MODEL[model_type]


def raise_when_operation_not_allowed(name: str, operation: str) -> None:
    """Verifies that the given operation is allowed for the given registered name.

    :raises: ValueError when the operation is not allowed.
    """
    if not _MODEL_REGISTRY_BY_NAME[name][f"can_{operation}"]:  # type: ignore
        raise ValueError(f"Operation '{operation}' not allowed for '{name}'")



from enum import Enum
from pydantic import BaseModel
from pydantic import ConfigDict

import typing


class Model(BaseModel):
    """
    Base Model for all Google Wallet Models.

    Sets a model_config for all Google Wallet Models that enforce that all attributes must be explicitly modeled, and trying to set an unknown attribute would raise an Exception.
    This Follows the Zen of Python (PEP 20) --> Explicit is better than implicit.
    """

    model_config = ConfigDict(
        extra="forbid",
        # use_enum_values=True,
    )


class WithIdModel(Model):
    """
    Model for Google Wallet models with an identifier.
    """

    id: str


def _snake_to_camel(snake_str: str) -> str:
    parts = snake_str.lower().split("_")
    return "".join(
        [(x.capitalize() if count != 0 else x) for count, x in enumerate(parts)]
    )


class CamelCaseAliasEnum(Enum):
    """Add an value alias in camelcase to the enum,
    given the value in snake-case.

    example: a enum like

    class FooExample(CamelCaseAliasEnum):
        FOO_BAR_BAZ = "FOO_BAR_BAZ"

    can be looked up like this:

    FooExample("fooBarBaz")

    """

    def __new__(cls: type["CamelCaseAliasEnum"], value: str) -> "CamelCaseAliasEnum":
        obj: "CamelCaseAliasEnum" = object.__new__(cls)
        obj._name_ = f"{cls.__name__} snake case literal"
        camel = _snake_to_camel(value)

        # create a second object with the camelcase name
        # creating an alias only does not work out since
        # pydantic checks for the value in the enum and not only the name
        camel_obj = object.__new__(cls)
        camel_obj._value_ = camel
        camel_obj._name_ = f"{cls.__name__} camel case alias"
        cls._value2member_map_[camel] = camel_obj
        cls._member_map_[camel] = camel_obj
        cls._member_names_.append(camel)
        return obj

    def __eq__(self, other: typing.Any | Enum) -> bool:
        """Allow comparison with the camelcase value.
        Take into account that UPPER_CASE and camelCase are equal
        """
        if not isinstance(other, Enum):
            other = self.__class__(other)
        if self.value == other.value:
            return True
        # this is not 100% correct, but should be good enough
        v1 = self.value.lower().replace("_", "")
        v2 = other.value.lower().replace("_", "")
        return v1 == v2



from ..bases import Model
from ..deprecated import DeprecatedKindFieldMixin
from pydantic import Field


# Attribute order as in Google's documentation to make future updates easier!
# last check: 2025-01-22


class TranslatedString(DeprecatedKindFieldMixin, Model):
    """
    see: https://developers.google.com/wallet/generic/rest/v1/LocalizedString#translatedstring
    """

    # inherits kind (deprecated)
    language: str | None = Field(default=None)
    value: str | None = Field(default=None)


class LocalizedString(DeprecatedKindFieldMixin, Model):
    """
    see: https://developers.google.com/wallet/generic/rest/v1/LocalizedString
    """

    # inherits kind (deprecated)
    translatedValues: list[TranslatedString] = Field(default_factory=list)
    defaultValue: TranslatedString



"""
This module contains models that are not directly related to passes but to other Google Wallet APIs
"""

from ..registry import register_model
from .bases import Model
from .bases import WithIdModel
from .datatypes.general import CallbackOptions
from .datatypes.message import Message
from .datatypes.smarttap import IssuerContactInfo
from .datatypes.smarttap import IssuerToUserInfo
from .datatypes.smarttap import Permission
from .datatypes.smarttap import SmartTapMerchantData
from .deprecated import DeprecatedKindFieldMixin
from .passes import generic
from .passes import retail
from .passes import tickets_and_transit
from pydantic import AnyHttpUrl
from pydantic import Field


# Attribute order as in Google's documentation to make future updates easier!
# last check: 2025-01-22


@register_model(
    "SmartTap",
    url_part="smartTap",
    can_read=False,
    can_update=False,
    can_list=False,
    can_message=False,
)
class SmartTap(DeprecatedKindFieldMixin, WithIdModel):
    """
    see: https://developers.google.com/wallet/generic/rest/v1/smarttap#resource:-smarttap
    """

    # inherits id
    merchantId: str
    infos: list[IssuerToUserInfo] | None = None


@register_model(
    "Issuer",
    url_part="issuer",
    resource_id="issuerId",
    can_message=False,
)
class Issuer(Model):
    """
    see: https://developers.google.com/wallet/generic/rest/v1/issuer
    """

    issuerId: str
    name: str
    contactInfo: IssuerContactInfo | None = None
    homepageUrl: AnyHttpUrl | None = None
    smartTapMerchantData: SmartTapMerchantData | None = None
    callbackOptions: CallbackOptions | None = None


@register_model(
    "Permissions",
    url_part="permissions",
    resource_id="issuerId",
    can_list=False,
    can_message=False,
)
class Permissions(Model):
    """
    see: https://developers.google.com/wallet/generic/rest/v1/permissions
    """

    issuerId: str | None = None
    permissions: list[Permission] = Field(default_factory=list)


class AddMessageRequest(Model):
    """
    see: https://developers.google.com/wallet/tickets/events/rest/v1/AddMessageRequest
    """

    message: Message | None = None


@register_model(
    "Jwt",
    url_part="jwt",
    can_read=False,
    can_update=False,
    can_list=False,
    can_message=False,
)
class JwtResource(Model):
    """
    see: https://developers.google.com/wallet/reference/rest/v1/jwt
         https://developers.google.com/wallet/tickets/events/rest/v1/jwt
         https://developers.google.com/wallet/generic/web/javascript-button#google-pay-api-for-passes-jwt
    """

    jwt: str


class Resources(Model):
    """
    see: https://developers.google.com/wallet/reference/rest/v1/jwt/insert#resources
         https://developers.google.com/wallet/tickets/events/rest/v1/jwt/insert#resources
    """

    eventTicketClasses: list[tickets_and_transit.EventTicketClass] | None = None
    eventTicketObjects: list[tickets_and_transit.EventTicketObject] | None = None
    flightClasses: list[tickets_and_transit.FlightClass] | None = None
    flightObjects: list[tickets_and_transit.FlightObject] | None = None
    giftCardClasses: list[retail.GiftCardClass] | None = None
    giftCardObjects: list[retail.GiftCardObject] | None = None
    loyaltyClasses: list[retail.LoyaltyClass] | None = None
    loyaltyObjects: list[retail.LoyaltyObject] | None = None
    offerClasses: list[retail.OfferClass] | None = None
    offerObjects: list[retail.OfferObject] | None = None
    transitClasses: list[tickets_and_transit.TransitClass] | None = None
    transitObjects: list[tickets_and_transit.TransitObject] | None = None
    genericClasses: list[generic.GenericClass] | None = None
    genericObjects: list[generic.GenericObject] | None = None


class JwtResponse(Model):
    """
    see: https://developers.google.com/wallet/tickets/events/rest/v1/jwt/insert
    """

    saveUri: AnyHttpUrl
    resources: Resources



from ..bases import Model
from ..datatypes.localized_string import LocalizedString
from ..datatypes.money import Money
from pydantic import Field
from pydantic import model_validator


# Attribute order as in Google's documentation to make future updates easier!
# last check: 2025-01-22


class LoyaltyPointsBalance(Model):
    """
    data-type,
    see: https://developers.google.com/wallet/retail/loyalty-cards/rest/v1/loyaltyobject#LoyaltyPointsBalance
    """

    string: str | None = None
    int_: int | None = Field(alias="int", serialization_alias="int", default=None)
    double: float | None = None
    money: Money | None = None

    @model_validator(mode="after")
    def check_one_of(self) -> "LoyaltyPointsBalance":
        given_values = [
            val
            for val in (self.string, self.int_, self.double, self.money)
            if val is not None
        ]
        if len(given_values) == 0:
            raise ValueError("One of string, int, double, or money must be set")
        if len(given_values) > 1:
            raise ValueError("Only one of string, int, double, or money must be set")
        return self


class LoyaltyPoints(Model):
    """
    data-type,
    see: https://developers.google.com/wallet/retail/loyalty-cards/rest/v1/loyaltyobject#LoyaltyPoints
    """

    label: str | None = None
    balance: LoyaltyPointsBalance | None = None
    localizedLabel: LocalizedString | None = None



from edutap.wallet_google.handlers.fastapi import router_callback
from fastapi import FastAPI
from fastapi.logger import logger

import os
import pathlib


app = FastAPI()
app.include_router(router_callback)


class LoggingCallbackHandler:
    """
    Implementation of edutap.wallet_google.protocols.CallbackHandler
    """

    async def handle(
        self,
        class_id: str,
        object_id: str,
        event_type: str,
        exp_time_millis: int,
        count: int,
        nonce: str,
    ) -> None:
        pathlib.Path(
            os.environ.get(
                "EDUTAP_WALLET_GOOGLE_EXAMPLE_CALLBACK_LOG_FILE", "./callback_log.txt"
            )
        )
        line = f'"{class_id}", "{object_id}", "{event_type}", "{exp_time_millis}", "{count}", "{nonce}"\n'
        logger.info(line)
        with open("callback_log.txt", "a") as file:
            file.write(line)



from ..bases import Model
from .enums import Action
from .enums import Role
from pydantic import AnyHttpUrl
from pydantic import EmailStr
from pydantic import Field


# Attribute order as in Google's documentation to make future updates easier!
# last check: 2025-01-22


class Permission(Model):
    """
    see: https://developers.google.com/wallet/generic/rest/v1/permissions#permission
    """

    emailAddress: EmailStr | None
    role: Role | None


class AuthenticationKey(Model):
    """
    see: https://developers.google.com/wallet/tickets/events/rest/v1/issuer#authenticationkey
    """

    id: int
    publicKeyPem: str


class SignUpInfo(Model):
    """
    see: https://developers.google.com/wallet/generic/rest/v1/smarttap#signupinfo
    """

    classId: str


class IssuerToUserInfo(Model):
    """
    see: https://developers.google.com/wallet/generic/rest/v1/smarttap#issuertouserinfo
    """

    action: Action = Action.ACTION_UNSPECIFIED
    url: AnyHttpUrl | None = Field(
        description="Currently not used, consider deprecating", default=None
    )
    value: str | None = None
    signUpInfo: SignUpInfo | None = None


class IssuerContactInfo(Model):
    """
    see: https://developers.google.com/wallet/generic/rest/v1/issuer#issuercontactinfo
    """

    name: str | None = None
    phone: str | None = None
    email: EmailStr | None = None
    alertsEmails: list[EmailStr] | None = None


class SmartTapMerchantData(Model):
    """
    see: https://developers.google.com/wallet/generic/rest/v1/issuer#smarttapmerchantdata
    """

    smartTapMerchantId: str | None = None
    authenticationKeys: list[AuthenticationKey] | None = None
